/* Do an 'equal-area' projection of the universe onto a plane.  This looks
   like the canonical IRAS sky maps */

#include <math.h>
#include <stdio.h>
#include <float.h>
#include <limits.h>
#include "Msgs.h"
#include "Assert.h"
#include "protos.h"
#include "Malloc.h"
#include "physics_n.h"
#include "fastflpt.h"
#include "vop.h"
#include "mpmy.h"
#include "gc.h"

#include "eaimage.h"

#if NDIM != 3
  # error this code only makes sense for 3-d simulations
#endif

struct accumulator *image;

static void Accumulate(int npix);
static void Output(const char *base, int npix);

void
eaimage(int npix, int nobj, body *btab, const char *base){
  body *bp, *btabend;
  float npix_2PI;
  int npixall;

  npixall = npix*((npix+1)/2);
  image = Calloc(npixall, sizeof(struct accumulator));
  
  npix_2PI = npix/(2.000001*M_PI);
  btabend = btab + nobj;
  for(bp=btab; bp<btabend; bp++){
    float rho, theta, phi, sintheta;
    float u, v;
    int i, j;
    int idx;
    float r2, r;
    float bweight, vr;

    /* Figure out which face we project onto by finding the largest
       absolute coordinate.  Then project intu (u,v) \in ([-1,1], [-1,1]) */
    phi = atan2(bp->pos[1], bp->pos[0]);
    rho = sqrtf_fast(bp->pos[1]*bp->pos[1] + bp->pos[0]*bp->pos[0]);
    theta = atan2(bp->pos[2], rho);
    r2 = Dot(bp->pos, bp->pos);
    r = sqrtf_fast(r2);
    sintheta = rho/r;
    /* u = phi * sintheta; */		/* range: [-pi, pi] */
    if( phi > 0 ){
      u = (phi - M_PI/2.0)*sintheta + M_PI/2.0;
    }else{
      u = (phi + M_PI/2.0)*sintheta - M_PI/2.0;
    }
    v = theta;			/* range: [-pi/2, pi/2] */
    /* Figure out which pixel to increment. */
    i = (u+M_PI)*npix_2PI;
    assert( i >= 0 && i < npix);
    j = (v+(M_PI/2.0))*npix_2PI;
    assert( j >= 0 && j < (npix+1)/2);
    idx = j*npix + i;

    /* Now accumulate whatever it is we are trying to sum about
       each pixel. */
    image[idx].mweight += bp->mass;
    bweight = bp->mass/r2;
    image[idx].bweight += bweight;
    vr = Dot(bp->pos, bp->vel)/r;
    image[idx].vrm += vr*bp->mass;
    image[idx].vrb += vr*bweight;
  }

  Accumulate(npixall);
  Output(base, npix);

  Free(image);
}

static void 
Accumulate(int npixall){
  /* An unrolled hypercube 'combine', with some data compression thrown
     in, and the final data only arriving in node 0! */
  int p;
  void *rle;
  int *rlep;
  int rlelen;
  int nempty, nfull;
  int nemptysum, nfullsum;
  int counting_empty;
  void *first_full;
  int doc, chan;
  MPMY_Comm_request req;

  rlelen = (npixall+1) * (sizeof(struct accumulator));/* gross overestimate */
  rle = Malloc(rlelen);
  doc = ilog2(MPMY_Nproc());
  if( (1<<doc) < MPMY_Nproc() )
    doc++;
  
  nfullsum = nemptysum = 0;
  for(chan=doc-1; chan>=0; chan--){
    int partner;
    int chanmask = 1<<chan;
    partner = MPMY_Procnum() ^ chanmask;

    if( chanmask & MPMY_Procnum() ){
      /* I am further from node 0.  'Compress' image and ship it out */
      counting_empty = 1;
      nempty = 0;
      nfull = 0;
      rlep = rle;
      for(p=0; p<npixall; p++){
	if( counting_empty ){
	  if( image[p].mweight == 0.F ){
	    nempty++;
	  }else{
	    counting_empty = 0;
	    first_full = &image[p];
	    nfull = 1;
	  }
	}else{
	  if( image[p].mweight == 0.F ){
	    *rlep++ = nempty;
	    *rlep++ = nfull;
	    memcpy(rlep, first_full, nfull*sizeof(struct accumulator));
	    rlep = (int *)( ((char *)rlep) + nfull*sizeof(struct accumulator));
	    Msgf(("Writing RLE pair:  %d empty, %d full\n", nempty, nfull));
	    nfullsum += nfull;
	    nemptysum += nempty;
	    nfull = 0;
	    nempty = 1;
	    counting_empty = 1;
	  }else{
	    nfull++;
	  }
	}
      }
      *rlep++ = nempty;
      *rlep++ = nfull;
      memcpy(rlep, first_full, nfull*sizeof(struct accumulator));
      rlep = (int *)( ((char *)rlep) + nfull*sizeof(struct accumulator) );
      nfullsum += nfull; nemptysum += nempty;
      Msgf(("Writing RLE pair:  %d empty, %d full\n", nempty, nfull));
      Msgf(("Finished writing %d full %d empty pixels\n", 
	     nfullsum, nemptysum));
      MPMY_Isend(rle, (char *)rlep - (char *)rle, partner, 3456, &req);
      MPMY_Wait(req, NULL);
      Msgf(("Sent %d bytes on chan %d\n", (char *)rlep - (char *)rle, chan));
      return;			/* !!! I'm DONE !!! */
    }else{
      /* I am closer to node 0.  Read data and add it to my copy of image */
      int npix;
      struct accumulator *localp, *inp;
      MPMY_Status mpmystat;
      
      if( partner > MPMY_Nproc() )
	continue;
      MPMY_Irecv(rle, rlelen, partner, 3456, &req);
      MPMY_Wait(req, &mpmystat);
      /* Decompress the data and add it into 'image' */
      Msgf(("Received %d bytes on chan %d\n", mpmystat.MPMY_COUNT, chan));
      npix = 0;
      rlep = rle;
      localp = image;
      while( npix < npixall ){
	nempty = *rlep++;
	nfull = *rlep++;
	localp += nempty;
	inp = (struct accumulator *)rlep;
	Msgf(("Reading RLE pair:  %d empty, %d full\n", nempty, nfull));
	npix += nempty + nfull;
	while( nfull-- ){
	  localp->mweight += inp->mweight;
	  localp->bweight += inp->bweight;
	  localp->vrm += inp->vrm;
	  localp->vrb += inp->vrb;
	  localp++;
	  inp++;
	}
	rlep = (int *)inp;
      }
      assert(((char*)rlep - (char*)rle) == mpmystat.MPMY_COUNT);
    }
  }

  /* Only processor 0 gets here.  */
  for(p=0; p<npixall; p++){
    if( image[p].bweight > 0. ){
      image[p].vrb /= image[p].bweight;
      image[p].vrm /= image[p].mweight;
    }
  }
}

/* How much processing do we want to do on output?  Probably not much.
   Fiddling with color maps, orientations, etc. is probably best left to
   post-processing. */
static void 
Output(const char *base, int npix){
  FILE *fp;
  char name[128];
  char header[1024];
  float mweightmax = -FLT_MAX, mweightmin = FLT_MAX;
  float bweightmax = -FLT_MAX, bweightmin = FLT_MAX;
  float vrmmax = -FLT_MAX, vrmmin = FLT_MAX;
  float vrbmax = -FLT_MAX, vrbmin = FLT_MAX;
  int npixall = npix * ((npix+1)/2);
  int p, hdrlen;
  struct accumulator *acc;
  
  if( MPMY_Procnum() != 0 )
    return;

  for(p=0; p<npixall; p++){
#define MINMAX(var) do{ \
  if( var##max < acc->var ) var##max = acc->var; \
  if( var##min > acc->var ) var##min = acc->var; }while(0)

    acc = &image[p];
    if( acc->mweight == 0. )
      continue;
    MINMAX(mweight);
    MINMAX(bweight);
    MINMAX(vrm);
    MINMAX(vrb);
  }

  sprintf(header, 
	  "# SDF\n"
	  "float mweightmax = %g;\n"
	  "float mweightmin = %g;\n"
	  "float bweightmax = %g;\n"
	  "float bweightmin = %g;\n"
	  "float vrmmax = %g;\n"
	  "float vrmmin = %g;\n"
	  "float vrbmax = %g;\n"
	  "float vrbmin = %g;\n"
	  "float x_pixels=%d;\n"
	  "float y_pixels=%d;\n"
	  "struct {\n"
	    "float mweight;\n"
	    "float bweight;\n"
	    "float vrm;\n"
	    "float vrb;\n"
	  "}[%d];\n"
	  "#SDF-EOH",
	  mweightmax, mweightmin,
	  bweightmax, bweightmin,
	  vrmmax, vrmmin,
	  vrbmax, vrbmin,
	  npix, (npix+1)/2, npix*((npix+1)/2));
  hdrlen = strlen(header);
  memset(header + hdrlen, '\f', sizeof(header) - hdrlen);
  header[sizeof(header)-1] = '\n';
  
  sprintf(name, "%s.ea", base);
  fp = fopen(name, "w");
  fwrite(header, 1, sizeof(header), fp);
  fwrite(image, sizeof(struct accumulator), npixall, fp);
  fclose(fp);
}      
